# Copyright (c) 2017-2018 M. Grady Saunders
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require_relative 'env'
require_relative 'patterns'
require_relative 'string'

module Schizm

# Structure to resolve chunks of source code.
class Chunk

  def initialize filename, label
    @filename = String.new filename
    @label = String.new label
    @share = Env.var[:share] if Env.var? :share
    @target = nil
    @blocks = Array.new
    @labels = Array.new
    @params = Hash.new
  end

  attr_reader :filename
  attr_reader :label
  attr_accessor :share
  attr_accessor :target
  attr_accessor :blocks

  # Boolean reader for +@share+.
  def share?
    case
    when (@share == true) then return true
    when (@share != true) then return false
    end
  end

  # Boolean reader for +@target+.
  def target?
    return false unless @target.is_a? ::String
    return false unless @target != ""
    return true
  end

  # Hash organizing active chunks by +@filename+ and +@label+.
  @@hash =
    Hash.new do |hash, filename|
    hash[filename] =
      Hash.new do |hash, label|
      hash[label] = Chunk.new filename, label
      end
    end

  # Hash reader.
  def self.hash
    @@hash
  end

  # Hash iteration interface.
  def self.each
    @@hash.each do |filename, hash|
      hash.each do |label, chunk|
        yield chunk
      end
    end
  end

  # Push a code block.
  def push block, label = nil
    block = String.new block
    label = String.new label if label
    @blocks.push block
    @labels.push label
  end

  # Find a chunk relative to +self+. That is, consider only
  # chunks defined in the same file, plus chunks shared with the
  # entire project.
  def find label
    Chunk.each do |chunk|
      if chunk.label == label and
        (chunk.filename == @filename or chunk.share?)
        return chunk
      end
    end
    nil
  end

  # Resolve chunk references to generate the final source string.
  def to_s *trace
    if trace.include? @label
      raise "Infinite recursion!"
    end
    total = String.new(@blocks.join "\n")
    total.trim!
    while true
      unless (total.sub! CHUNK_LABEL do
          if (chunk = find $1)
            replace = chunk.to_s *trace, @label
            replace.gsub /(?<=\n)/, " " * String.new($`).find_indent
          else
            "/* " + $1 + " ... */"
          end
          end)
        break
      end
    end
    if target?
      if Env.var? :guard and @target =~ HEADER_PATH
        guardify = String.new(@target).guardify
        if Env.var? :title
          guardify.prepend "_"
          guardify.prepend String.new(Env.var[:title]).guardify
        end
        total.prepend "\#define #{guardify}\n\n"
        total.prepend "\#ifndef #{guardify}\n"
        total.concat "\n\n"
        total.concat "\#endif /* \#ifndef #{guardify} */"
      end
      if File.file? "LICENSE"
        total.prepend "\n"
        total.prepend String.new(File.read("LICENSE")).comment
      end
    end
    total
  end

  # Identifier for +block+.
  def id block = @blocks.size
    String.new "#{@label.slugify}-#{block}"
  end

end # class Chunk

end # module Schizm
