# Copyright (c) 2017-2018 M. Grady Saunders
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require 'etc'
require 'optparse'
require 'pathname'
require 'tempfile'
require 'sass'
require_relative 'colors'

module Schizm

module Env

  # Global variable hash.
  @@var = {}

  # Global variable hash getter.
  def self.var
    return @@var
  end

  # +true+ if +@@var[key]+ is neither +nil+ nor +false+.
  def self.var? key
    return true if @@var.has_key?(key) and
      @@var[key] != nil and @@var[key] != false
    return false
  end

  # Current user's name.
  def self.thisuser
    return Etc.getpwnam(Etc.getlogin).gecos.split(/,/).first
  end

  # Current user's working directory.
  def self.thisdirname
    return File.basename(Dir.pwd)
  end

  # Current project title.
  def self.title
    return var[:title] if var? :title
    return thisdirname
  end

  # Current project brief.
  def self.brief
    return var[:brief] if var? :brief
    return ""
  end

  # Current project author.
  def self.author
    return var[:author] if var? :author
    return thisuser
  end

  # Current project year.
  def self.year
    return var[:year] if var? :year
    return Time.new.year.to_s
  end

  # Schizm install directory.
  def self.home
    path = File.dirname __FILE__    # home/lib/schizm
    path = File.dirname path        # home/lib
    path = File.dirname path        # home
    return File.absolute_path path
  end

  # Load resource from Schizm install directory.
  def self.resource *args
    return File.read File.join(home, "res", *args)
  end

  # Prompt the user with a yes/no question.
  def self.yes? string, backup
    backup = backup.downcase[0]
    print "#{string} [Y/n] " if backup == "y"
    print "#{string} [y/N] " if backup == "n"
    answer = gets.strip.downcase[0]
    return answer == "y" if answer != ""
    return backup == "y"
  end

  # Write +string+ to +target+. If +target+ is an existing file and
  # +var[:rewrite]+ is not set, prompts the user for permission to rewrite.
  def self.write target, string
    unless File.file? target and 
        not var? :rewrite and 
        not yes? "Rewrite '#{target}'?", "n"
      puts "Writing '#{target}'" if var? :verbose
      file = File.open target, "wb"
      file.write string.to_s
      file.close
    end
  end

  # Delete +path+ if it exists. If +path+ is a directory, recursively
  # delete all subdirectories.
  def self.delete path
    if File.exist? path
      puts "Deleting #{path}" if var? :verbose
      if File.file? path
        File.delete path
      elsif File.directory? path
        Dir.foreach path do |subpath|
          next if subpath == '.'
          next if subpath == '..'
          self.delete File.join(path, subpath)
        end
      end
    end
  end

  # Create missing directories in +path+.
  def self.blaze path
    dir = ''
    path = File.dirname path
    path = File.absolute_path path
    dirs = path.split File::SEPARATOR
    dirs.each do |sub|
      dir = File.join dir, sub
      unless Dir.exist? dir
        puts "Making '#{dir}'" if var? :verbose
        Dir.mkdir dir
      end
    end
  end

  # Return path +dest+ relative to +from+.
  def self.path_from from, dest
    return Pathname.new(dest).relative_path_from(Pathname.new(from)).to_s
  end

  def self.docs_path path
    return "/docs/#{path_from var[:output_doc], path}"
  end

  module Opts

    def self.init
      return OptionParser.new do |optparse|
        optparse.banner = "Usage: schizm init [OPTIONS]"
        optparse.on "-v", "--verbose", "Verbose messages?" do
          Env.var[:verbose] = true
        end
        optparse.on "-r", "--rewrite", "Rewrite files without asking?" do
          Env.var[:rewrite] = true
        end
        optparse.on "-t", "--title STR", "Set project title." do |str|
          Env.var[:title] = str
        end
        optparse.on "-b", "--brief STR", "Set project brief." do |str|
          Env.var[:brief] = str
        end
        optparse.on "-a", "--author STR", "Specify project author(s)." do |str|
          Env.var[:author] = str
        end
        optparse.on "-y", "--year STR", "Specify project year(s)." do |str|
          Env.var[:year] = str
        end
        license_enum = [
          "BSD-2-Clause",
          "BSD-3-Clause",
          "GPL-2.0",
          "GPL-3.0",
          "LGPL-2.1",
          "LGPL-3.0",
          "MIT",
          "MPL-2.0"
        ]
        license_help = <<LIC
Select license, none by default.

\tENUM                License
\t"BSD-2-Clause" .... the 2-clause BSD License
\t"BSD-3-Clause" .... the 3-clause BSD License
\t"GPL-2.0" ......... the GNU General Public License v2.0
\t"GPL-3.0" ......... the GNU General Public License v3.0
\t"LGPL-2.1" ........ the GNU Lesser General Public License v2.1
\t"LGPL-3.0" ........ the GNU Lesser General Public License v3.0
\t"MIT" ............. the MIT License
\t"MPL-2.0" ......... the Mozilla Public License v2.0

LIC
        optparse.on "-l", "--license ENUM",
            license_enum,
            license_help do |str|
          Env.var[:license] = str
        end
        color_help = <<COL

\tENUM                Basic color
\t"red" ............. F4 43 36
\t"pink" ............ E9 1E 63
\t"purple" .......... 9C 27 B0 *** Default primary
\t"deep-purple" ..... 67 3A B7
\t"indigo" .......... 3F 51 B5
\t"blue" ............ 21 96 F3
\t"light-blue" ...... 03 A9 F4
\t"cyan" ............ 00 BC D4
\t"teal" ............ 00 96 88 *** Default secondary
\t"green" ........... 4C AF 50
\t"light-green" ..... 8B C3 4A
\t"lime" ............ CD DC 39
\t"yellow" .......... FF EB 3B
\t"amber" ........... FF C1 07
\t"orange" .......... FF 98 00
\t"deep-orange" ..... FF 57 22

\tRefer to https://material.io/guidelines/style/color.html

COL
        optparse.on "-p", "--primary-color ENUM", COLORS.keys,
          "Specify primary color." do |str|
          Env.var[:primary_color] = str
        end
        optparse.on "-s", "--secondary-color ENUM", COLORS.keys,
          "Specify secondary color.", color_help do |str|
          Env.var[:secondary_color] = str
        end
        optparse.on_tail "-h", "--help", "Display this help." do
          puts optparse
          exit 0
        end
      end
    end

  end # module Opts

  def self.init
    dirnames = [
      "src",
      "zrc",
      "docs",
      "docs/assets",
      "docs/assets/css",
      "docs/assets/fonts",
      "docs/assets/js",
      "docs/_includes",
      "docs/_layouts",
      "docs/_plugins",
      "docs/_posts",
      "docs/_docs",
      "docs/_site"
    ]
    dirnames.each do |dirname|
      unless File.directory? dirname
        Dir.mkdir dirname
      end
    end
    license = String.new
    license_content = String.new
    if var? :license
      license << var[:license]
      license_content << "Copyright (c) #{year} #{author}\n\n"
      license_content << "#{resource("license", license)}"
      write "LICENSE", license_content
    end
    write "README.md", <<README
# #{title}
README
    write "Makefile", <<MAKEFILE
SHELL := /bin/bash
MKDIR := mkdir -p
RMDIR := rm -r -f

SCHIZM := schizm
ZFLAGS := --input zrc
ZFLAGS += --output-src src
ZFLAGS += --output-doc docs/_docs
ZFLAGS += --rewrite
ZFLAGS += --verbose
ZFLAGS += --title "#{title}"
ZFLAGS += --brief "#{brief}"
ZFLAGS += --author "#{author}"
ZFLAGS += --year "#{year}"

doc:
\t@$(SCHIZM) build --only-doc $(ZFLAGS)

src:
\t@$(SCHIZM) build --only-src $(ZFLAGS)

.PHONY: doc

.PHONY: src

clean:
\t@$(RMDIR) src docs/_docs docs/_site
\t@$(MKDIR) src docs/_docs docs/_site

.PHONY: clean
MAKEFILE
    write ".gitignore", <<GITIGNORE
*.swp
*.tmp
docs/_site
GITIGNORE
    write "docs/_config.yml", <<YMLCONFIG
source: .
destination: _site
plugins_dir: _plugins
layouts_dir: _layouts
includes_dir: _includes
collections:
  posts:
    output: true
  docs:
    output: true
encoding: "utf-8"
markdown: kramdown
markdown_ext: "markdown,md"
author: "#{author}"
title: "#{title}"
brief: "#{brief}"
YMLCONFIG
    assets = {
      "schizm.liquid" => "docs/_layouts",
      "schizm.js" => "docs/assets/js",
      "fonts/Fira-Code.css" => "docs/assets/fonts",
      "fonts/Fira-Code-Light.otf" => "docs/assets/fonts",
      "fonts/Fira-Code-Regular.otf" => "docs/assets/fonts",
      "fonts/Fira-Code-Medium.otf" => "docs/assets/fonts",
      "fonts/Fira-Code-Bold.otf" => "docs/assets/fonts",
      "fonts/Fira-Code-LICENSE" => "docs/assets/fonts",
      "fonts/Genericons-Neue.css" => "docs/assets/fonts",
      "fonts/Genericons-Neue.eot" => "docs/assets/fonts",
      "fonts/Genericons-Neue.ttf" => "docs/assets/fonts",
      "fonts/Genericons-Neue.woff2" => "docs/assets/fonts",
      "fonts/Genericons-Neue-LICENSE" => "docs/assets/fonts",
      "fonts/Social-Logos.css" => "docs/assets/fonts",
      "fonts/Social-Logos.eot" => "docs/assets/fonts",
      "fonts/Social-Logos.ttf" => "docs/assets/fonts",
      "fonts/Social-Logos.woff2" => "docs/assets/fonts",
      "fonts/Social-Logos-LICENSE" => "docs/assets/fonts"
    }
    assets.each do |from, to|
      write File.join(to, File.basename(from)), resource(from)
    end
    color1 = "purple"
    color2 = "teal"
    color1 = var[:primary_color].downcase if var? :primary_color
    color2 = var[:secondary_color].downcase if var? :secondary_color
    sass_source = <<SASS
#{Schizm.sass_colors color1, "primary"}
#{Schizm.sass_colors color2, "secondary"}
#{resource "scss/fonts.scss"}
#{resource "scss/mixin.scss"}
#{resource "scss/style.scss"}
SASS
    sass_options = {
      :syntax => :scss,
      :sourcemap => :none,
      :style => :nested,
      :cache => false
    }
    write "docs/assets/css/schizm.css", Sass::Engine.new(sass_source, sass_options).render
  end

  module Opts

    def self.build
      Env.var[:input] = "zrc"
      Env.var[:output_doc] = "./docs/_docs"
      Env.var[:output_src] = "./src"
      Env.var[:guard] = true
      return OptionParser.new do |optparse|
        optparse.banner = "Usage: schizm build [OPTIONS]"
        optparse.on "-v", "--verbose", "Verbose messages?" do
          Env.var[:verbose] = true
        end
        optparse.on "-r", "--rewrite", "Rewrite files without asking?" do
          Env.var[:rewrite] = true
        end
        optparse.on "-t", "--title STR", "Set project title." do |str|
          Env.var[:title] = str
        end
        optparse.on "-b", "--brief STR", "Set project brief." do |str|
          Env.var[:brief] = str
        end
        optparse.on "-a", "--author STR", "Specify project author(s)." do |str|
          Env.var[:author] = str
        end
        optparse.on "-y", "--year STR", "Specify project year(s)." do |str|
          Env.var[:year] = str
        end
        optparse.on "--input STR",
          "Where to look for input files." do |str|
          Env.var[:input] = str
        end
        optparse.on "--output-doc STR",
          "Where to write documentation files." do |str|
          Env.var[:output_doc] = str
        end
        optparse.on "--output-src STR",
          "Where to write source files." do |str|
          Env.var[:output_src] = str
        end
        optparse.on "--only-doc", "Build only documention files?" do
          Env.var[:only_doc] = true
        end
        optparse.on "--only-src", "Build only source files?" do
          Env.var[:only_src] = true
        end
        optparse.on "--[no-]guard", "Guard output headers?" do |ans|
          Env.var[:guard] = ans
        end
        optparse.on "--[no-]share", "Share chunks by default?" do |ans|
          Env.var[:share] = ans
        end
        optparse.on_tail "-h", "--help", "Display this help." do
          puts optparse
          exit 0
        end
      end
    end

  end # module Opts

  def self.build
    zrc_root = var[:input]
    doc_root = var[:output_doc]
    src_root = var[:output_src]
    zrc_globs = []
    zrc_globs.push File.join(zrc_root, "*.zm")
    zrc_globs.push File.join(zrc_root, "*/*.zm")
    zrc_paths = Dir.glob zrc_globs
    raise "No input files." if zrc_paths.empty?
    delete doc_root if var? :clean_doc
    delete src_root if var? :clean_src
    blaze doc_root
    blaze src_root
    zrc_paths.each do |path|
      filename = path_from zrc_root, path
      filename = File.join doc_root, filename.chomp(".zm").concat(".html")
      input = File.read path
      Parse.hash[filename] =
      Parse.page_elems filename, input
    end
    unless Env.var? :only_src
      Parse.hash.each do |filename, parse|
        blaze filename
        write filename, <<PARSE
---
layout: schizm
author: "#{author}"
#{parse.guess_title}
#{parse.guess_description}
#{parse.guess_parts}
---
#{parse.to_s}
PARSE
      end
    end
    unless Env.var? :only_doc
      Chunk.each do |chunk|
        if chunk.target?
          blaze filename = File.join(src_root, chunk.target)
          write filename, chunk
        end
      end
    end
  end

  module Opts

    def self.vim which
      return OptionParser.new do |optparse|
        optparse.banner = "Usage: schizm vim-#{which} [OPTIONS]"
        optparse.on "-v", "--verbose", "Verbose messages?" do
          Env.var[:verbose] = true
        end
        optparse.on "-r", "--rewrite", "Rewrite files without asking?" do
          Env.var[:rewrite] = true
        end
        optparse.on "-p", "--path STR", "Specify Vim path." do |str|
          Env.var[:vimhome] = str
        end
        optparse.on_tail "-h", "--help", "Display this help." do
          puts optparse
          exit 0
        end
      end
    end

    def self.vim_install
      return vim "install"
    end

    def self.vim_uninstall
      return vim "uninstall"
    end

  end

  def self.vim_install
    vimhome = "~/.vim"
    vimhome = var[:vimhome] if var? :vimhome
    vimhome = File.absolute_path File.expand_path(vimhome)
    assets = {
      "vim/ftdetect/schizm.vim" => File.join(vimhome, "ftdetect"),
      "vim/ftplugin/schizm.vim" => File.join(vimhome, "ftplugin"),
      "vim/indent/schizm.vim" => File.join(vimhome, "indent"),
      "vim/syntax/schizm.vim" => File.join(vimhome, "syntax")
    }
    assets.each do |from, to|
      blaze File.join(to, File.basename(from))
      write File.join(to, File.basename(from)), resource(from)
    end
  end

  def self.vim_uninstall
    vimhome = "~/.vim"
    vimhome = var[:vimhome] if var? :vimhome
    vimhome = File.absolute_path File.expand_path(vimhome)
    assets = [
      File.join(vimhome, "ftdetect/schizm.vim"),
      File.join(vimhome, "ftplugin/schizm.vim"),
      File.join(vimhome, "indent/schizm.vim"),
      File.join(vimhome, "syntax/schizm.vim")
    ]
    assets.each do |asset|
      delete asset
    end
  end

end # module Env

end # module Schizm
