# Copyright (c) 2017-2018 M. Grady Saunders
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require_relative 'string'

module Schizm

class Markup

  def initialize tag = nil
    @tag = tag
    @attrs = {}
    @elems = []
  end

  attr_reader :tag
  attr_accessor :attrs
  attr_accessor :elems

  # Attribute accessor.
  def [] key
    return @attrs[key]
  end

  # Attribute accessor.
  def []= key, value
    return @attrs[key] = value
  end

  # Attribute +name+ or +nil+ if +name+ does not exist.
  def get_attr name
    return @attrs[name] if @attrs.has_key? name
    return nil
  end

  # Wrapper for +@attrs.has_key?+.
  def has_attr? name 
    return @attrs.has_key? name
  end

  # Add attributes from hash +pairs+.
  def add_attrs pairs
    pairs.each do |key, value|
      @attrs[key] = value 
    end
    return self
  end

  # Add element(s).
  def << param
    case param
    when nil
      # Ignore silently.
    when ::Array
      for entry in param
        self << entry
      end
    when ::Integer
      self << '&#' << param.to_s << ';'
    when ::String
      target = @elems.last
      target = @elems unless target.is_a? ::String
      target << param
    else
      @elems << param
    end
    return self
  end

  # Wrapper for +self << param+.
  def add_elems param
    return self << param
  end

  # Wrapper for +@elems.last+.
  def last
    return @elems.last
  end

  def yield_each *tags, &block
    if @tag and tags.include? @tag.to_s
      yield self
    end
    for elem in @elems
      elem.yield_each *tags, &block if elem.is_a? Markup
    end
  end

  def find_first_of *tags, &block
    if @tag and tags.include? @tag.to_s
      return self if not block or (block.call(self) == true)
    end
    for elem in @elems
        temp = elem.find_first_of *tags, &block if elem.is_a? Markup
        return temp if temp
    end
    nil
  end

  def to_plain_text
    text = String.new
    @elems.each do |elem|
      case elem
      when Markup 
        if elem.tag == "script" and 
          (elem["type"] == "math/tex; mode=inline" or
           elem["type"] == "math/tex; mode=display")
          temp = elem.elems[0].dump
          text << String.new(temp[1..(temp.size-2)]).html
        else
          text << elem.to_plain_text
        end
      else
        text << elem.to_s if elem.respond_to? :to_s
      end
    end
    text
  end

  def to_s
    s = String.new
    unless @tag == nil
      s << "<" << @tag
      @attrs.each do |k, v|
        if k and k.to_s != ""
          s << " #{k.to_s}" if v == nil
          s << " #{k.to_s}=\"#{v.to_s}\"" if v != nil and v.to_s != ""
        end
      end
      s << ">"
    end
    @elems.each do |elem|
      s << elem.to_s if elem.respond_to? :to_s
    end
    unless @tag == nil
      s << "</" << @tag << ">" 
    end
    s
  end

  def guess_title
    elem = find_first_of "h1" do |temp|
      true if temp.has_attr? "class" and temp["class"] == "title"
    end
    return "title: \"#{elem.to_plain_text}\"" if elem
    return "title: \"\""
  end

  def guess_description
    elem = find_first_of "p"
    text = elem.to_plain_text if elem
    return "description: \"#{text.truncate 80}\"" if text
    return "description: \"\""
  end

  def guess_parts
    tab = "  "
    parts = "parts:\n"
    yield_each "h2" do |elem|
      if elem.get_attr("class") == "title" and
         elem.get_attr("id") != nil
        parts.concat "#{tab}-\n"
        parts.concat "#{tab}#{tab}id: \"#{elem.get_attr "id"}\"\n"
        parts.concat "#{tab}#{tab}title: \"#{elem.to_plain_text}\"\n"
      end
    end
    parts
  end

end # class Markup

end # module Schizm
