# Copyright (c) 2017-2018 M. Grady Saunders
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module Schizm

  # Header path, extension put in match group 1.
  HEADER_PATH = /^.*\.(h|hh|hpp|hxx)$/

  # Source path, extension put in match group 1.
  SOURCE_PATH = /^.*\.(c|cc|cpp|cxx|inl|tcc)$/

  # Space sequence.
  SPACE = /\p{Space}+/u

  # Indent level.
  INDENT = /^[ ]{4}/u

  # Empty line.
  EMPTY_LINE = /^\n/u

  # Empty lines.
  EMPTY_LINES = /#{EMPTY_LINE}*/u

  # Explicit end-of-block marker.
  EOB_MARKER = /^\^\n/u

  # Explicit or implicit end of simple block.
  END_SIMPLE = /^\^?\n|\z/u

  # Explicit or implicit end of indent block.
  END_INDENT = /\n(?=[ ]{0,3}\p{Graph})|\z/u

  # Setext heading.
  HEADING_SETEXT = /^[ ]{0,3}(\p{Graph}.*)\n(===+|---+)\n/u

  # Atx heading.
  HEADING_ATX = /^[ ]{0,3}(\#{1,6})(.+)\n/u

  # Horizontal rule.
  HRULE = /^(?:\*[ ]?){3,}(?:\n|\z)/u

  # 1. A character region escaped by inline code delimiters,
  # 2. a character region escaped by inline math delimiters, or
  # 3. a character which is not a backslash or a closing brace.
  #
  NOT_CLOSE_BRACE = /`(?:[^\\`]*(?:\\.)*)*`|\$(?:[^\\$]*(?:\\.)*)*\$|[^\\}]/um

  # 1. A character region escaped by inline code delimiters,
  # 2. a character region escaped by inline math delimiters, or
  # 3. a character which is not a backslash or a closing bracket.
  #
  NOT_CLOSE_BRACK = /`(?:[^\\`]*(?:\\.)*)*`|\$(?:[^\\$]*(?:\\.)*)*\$|[^\\\]]/um

  # 1. A character region escaped by inline code delimiters,
  # 2. a character region escaped by inline math delimiters, or
  # 3. a character which is not a backslash or a closing parenthesis.
  #
  NOT_CLOSE_PAREN = /`(?:[^\\`]*(?:\\.)*)*`|\$(?:[^\\$]*(?:\\.)*)*\$|[^\\)]/um

  # Matches until an unescaped closing brace.
  DELIM_BRACES = /#{NOT_CLOSE_BRACE}*(?:\\.#{NOT_CLOSE_BRACE}*)*/um

  # Matches until an unescaped closing bracket.
  DELIM_BRACKS = /#{NOT_CLOSE_BRACK}*(?:\\.#{NOT_CLOSE_BRACK}*)*/um

  # Matches until an unescaped closing parenthesis.
  DELIM_PARENS = /#{NOT_CLOSE_PAREN}*(?:\\.#{NOT_CLOSE_PAREN}*)*/um

  # HTML entity.
  HTML_ENTITY = /&(?:\w+|\#\d+|\#x\h+);/

  # HTML special characters.
  HTML_ESCAPE = /[\p{^ASCII}&<>]/u

  # Character escape, put in match group 1.
  CHAR_ESCAPE = /\\(.)/u

  # Link definition, content put in match group 1, params put in match group 2.
  LINK_DEF = /^&:\[(#{DELIM_BRACKS})\][ ]*(.*)[ ]*\n/u

  # Link reference, content put in match group 1, params put in match group 2.
  LINK_REF_PARENS = /&\[(#{DELIM_BRACKS})\]\((#{DELIM_PARENS})\)/um

  # Link reference, content put in match group 1, params put in match group 2.
  LINK_REF_BRACKS = /&\[(#{DELIM_BRACKS})\](?:\[(#{DELIM_BRACKS})\])?/um

  # Image definition, entry put in match group 1, params put in match group 2.
  IMAGE_DEF = /^!:\[(#{DELIM_BRACKS})\][ ]*(.*)[ ]*\n/u

  # Image reference, alt put in match group 1, params put in match group 2.
  IMAGE_REF_PARENS = /!\[(#{DELIM_BRACKS})\]\((#{DELIM_PARENS})\)/u

  # Image reference, alt put in match group 1, params put in match group 2.
  IMAGE_REF_BRACKS = /!\[(#{DELIM_BRACKS})\](?:\[(#{DELIM_BRACKS})\])?/u

  # Center math, content put in match group 1.
  CENTER_MATH = /\$\$((?:[^\\$]*(?:\\.)*)*)\$+/um

  # Markup attributes for +CENTER_MATH+.
  CENTER_MATH_ATTRS = {'type' => 'math/tex; mode=display'}

  # Inline math, content put in match group 1.
  INLINE_MATH = /\$((?:[^\\$]*(?:\\.)*)*)\$/um

  # Markup attributes for +INLINE_MATH+.
  INLINE_MATH_ATTRS = {'type' => 'math/tex; mode=inline'}

  # Inline code, content put in match group 1.
  INLINE_CODE = /[`]((?:[^\\`]*(?:\\.)*)*)[`]/um

  # Inline emphasis, depth put in match group 1, content put in match group 2.
  INLINE_EMPH = /([*]+)((?:[^\\*]*(?:\\.)*)*)[*]+/um

  # Inline subscript, content put in match group 1.
  INLINE_SUB = /_{(#{DELIM_BRACES})}/um

  # Inline superscript, content put in match group 1.
  INLINE_SUP = /\^{(#{DELIM_BRACES})}/um

  # Inline insert, content put in match group 1.
  INLINE_INS = /\+{(#{DELIM_BRACES})}/um

  # Inline delete, content put in match group 1.
  INLINE_DEL = /-{(#{DELIM_BRACES})}/um

  # Chunk label, content put in match group 1.
  CHUNK_LABEL = /@\[(#{DELIM_BRACKS})\]/um

  # Chunk block, content put in match group 1.
  CHUNK_BLOCK = /\((#{DELIM_PARENS})\)/um

  # Enter quote block.
  ENTER_QUOTE = /^[>][ ]{3}/u 

  # Enter aside block.
  ENTER_ASIDE = /^[~][ ]{3}/u 

  # Enter alert block.
  ENTER_ALERT = /^[!][ ]{3}/u 

  # Enter bug block.
  ENTER_BUG = /^[?][ ]{3}/u 

  # Enter olist element.
  ENTER_OLIST = /^\d\.[ ]{2}|\d{2}\.[ ]/u 

  # Enter ulist element.
  ENTER_ULIST = /^[-][ ]{3}|[-]{2}[ ]{2}/u 

  # Enter tlist element, type put in match group 1.
  ENTER_TLIST = /^\[([x ])\][ ]/u 

  # Enter dlist element, content put in match group 1.
  ENTER_DLIST = /^(\p{Graph}.*?)::/u 

  # Enter code block.
  ENTER_CODE = /^[ ]{4,}(?=\p{Graph})/u

  # Enter text block.
  ENTER_TEXT = /^[ ]{0,3}(?=\p{Graph})/u

  # Special ASCII sequences.
  GLYPH = [
    [/\(tm\)/ui, 8482], # Trademark.
    [/\(c\)/ui, 9400], # Copyright.
    [/\(r\)/ui, 174], # Registered.
    [/\+-/u, 177], # Plus minus.
    [/==/u, 8801], # Identical.
    [/=~/u, 8773], # Congruent.
    [/!=/u, 8800], # Not equal.
    [/<=/u, 8804], # Less equal.
    [/>=/u, 8805], # Greater equal.
    [/~=/u, 8776], # Approximately.
    [/--/u, 8211], # En dash.
    [/---/u, 8212], # Em dash.
    [/\.{3}/u, 8230], # Ellipsis.
    [/<--/u, 8592], # Left single arrow.
    [/<==/u, 8656], # Left double arrow.
    [/-->/u, 8594], # Right single arrow.
    [/==>/u, 8658], # Right double arrow.
    [/<-->/u, 8596], # Horizontal single arrow.
    [/<==>/u, 8660], # Horizontal double arrow.
    [/<~~/u, 8668], # Left squiggle arrow.
    [/~~>/u, 8669], # Right squiggle arrow.
    [/<<--/u, 8606], # Left two headed arrow.
    [/-->>/u, 8608], # Right two headed arrow.
    [/<--</u, 8610], # Left arrow with tail.
    [/>-->/u, 8611], # Right arrow with tail.
    [/\|<--/u, 8676], # Left arrow to bar.
    [/-->\|/u, 8677], # Right arrow to bar.
    [/<---/u, 10229], # Left long single arrow.
    [/<===/u, 10232], # Left long double arrow.
    [/--->/u, 10230], # Right long single arrow.
    [/===>/u, 10233], # Right long double arrow.
    [/<--->/u, 10231], # Horizontal long single arrow.
    [/<===>/u, 10234], # Horizontal long double arrow.
    [/<---\|/u, 10235], # Left long single arrow from bar.
    [/<===\|/u, 10237], # Left long double arrow from bar.
    [/\|--->/u, 10236], # Right long single arrow from bar.
    [/\|===>/u, 10238], # Right long double arrow from bar.
    [/<-\/-/u, 8602], # Left single arrow with stroke.
    [/<=\/=/u, 8653], # Left double arrow with stroke.
    [/<-\|-/u, 8695], # Left single arrow with vertical bar.
    [/<=\|=/u, 10498], # Left double arrow with vertical bar.
    [/-\/->/u, 8603], # Right single arrow with stroke.
    [/=\/=>/u, 8655], # Right double arrow with stroke.
    [/-\|->/u, 8696], # Right single arrow with vertical bar.
    [/=\|=>/u, 10499], # Right double arrow with vertical bar.
    [/<-\/->/u, 8622], # Horizontal single arrow with stroke.
    [/<-\|->/u, 8697], # Horizontal single arrow with vertical bar.
    [/<=\|=>/u, 10500], # Horizontal double arrow with vertical bar.
    [/<--\|/u,  8612], # Left single arrow from bar.
    [/<==\|/u, 10502], # Left double arrow from bar.
    [/\|-->/u,  8614], # Right single arrow from bar.
    [/\|==>/u, 10503]  # Right double arrow from bar.
  ]

  # C++ string.
  HICPPSTR = /(?:u8?|[ULR])?(["'])((?:[^\\\1]*(?:\\.)*)*)\1/u

  # C++ number.
  HICPPNUM = /[+-]?(?:(?:(?:(?:\.\d+|\d+\.\d*|\d+(?=e))(?:e[+-]?\d+)?)|0x1\.\h+p[+-]\d+)(?:i?[fl]|_\w+)?|(?:0b[01]+|0[0-7]+|[1-9]\d*|0(?!\d)|0x\h+)(?:ul?l?|ll?|_\w+)?)/i

  # C++ syntax highlighting.
  HICPP = [
    ["hi-comment", /\/\/.*$/u],
    ["hi-comment", /\/\*.*?\*\//um],
    ["hi-preproc", /\#(?:ifn?def|endif|if|else|elif|define|pragma|include)\b.*?(?<=[^\\])(?:\n|\z)/um],
    ["hi-type", /(?:auto|bool|char|short|long|int|(?:un)?signed|float|double|void)\b/u],
    ["hi-type", /(?:size|ptrdiff|char(?:16|32)|wchar|u?int(?:_least|_fast)?(?:8|16|32|64))_t\b/u],
    ["hi-type-spec", /(?:constexpr|const|volatile|static|mutable|thread_local|register|extern|inline|noexcept)\b/u],
    ["hi-flow", /(?:new|delete|do|while|for|if|else|try|throw|catch|switch|case|default|break|continue|goto|return)\b/u],
    ["hi-op", /(?:(?:and|or|xor|not)(?:_eq)?|bitand|bitor|compl|(?:(?:const|dynamic|static|reinterpret)_cast))\b/u],
    ["hi-value", /(?:true|false|this|nullptr|NULL|#{HICPPSTR}|#{HICPPNUM})\b/u],
    ["hi-other", /(?:alignas|alignof|sizeof|decltype|template|typename|typedef|using|concept|requires|operator)\b/u],
    ["hi-other", /(?:namespace|class|struct|enum|public|protected|private|friend|final|virtual|override)\b/u],
    ["hi-ident", /[a-zA-Z]\w*/u],
    [nil, /[^\w\s]+/um]
  ]

end # module Schizm
