# Copyright (c) 2017-2018 M. Grady Saunders
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module Schizm

# Extend the default +String+ class.
class String < ::String

  # Erase occurances of +Regexp.union(*args)+.
  def erase! *args
    gsub! Regexp.union(*args), ""
  end

  # Erase occurances of +Regexp.union(*args)+.
  def erase *args
    gsub Regexp.union(*args), ""
  end

  # Replace sequences of +\p{^Word}+ by an underscore.
  def wordify!
    gsub! /\p{^Word}+/u, '_'
  end

  # Replace sequences of +\p{^Word}+ by an underscore.
  def wordify
    gsub /\p{^Word}+/u, '_'
  end

  # Make suitable string for slug.
  def slugify
    gsub(/^\p{^Word}+/u, "")
      .gsub(/\p{^Word}+$/u, "")
      .gsub(/\p{^Word}+/u, '-').downcase
  end

  # Make suitable string for include guard.
  def guardify
    strip.wordify.upcase
  end

  # Encode non-ASCII characters and '&', '<', and '>' as HTML entities.
  def html!
    gsub! /[&<>\p{^ASCII}]/u do |char|
      '&#' + char.codepoints[0].to_s + ';'
    end
  end

  # Encode non-ASCII characters and '&', '<', and '>' as HTML entities.
  def html
    gsub /[&<>\p{^ASCII}]/u do |char|
      '&#' + char.codepoints[0].to_s + ';'
    end
  end

  # Remove blanks preceding newlines.
  def trim_blank!
    gsub! /\p{Blank}+\n/u, "\n"
  end

  # Remove blanks preceding newlines.
  def trim_blank
    gsub /\p{Blank}+\n/u, "\n"
  end

  # Remove blanks preceding newlines, remove trailing
  # spaces, remove newlines succeeding empty lines, and
  # remove leading newlines.
  def trim!
    gsub! /\p{Blank}+\n/u, "\n"
    gsub! /\p{Space}+\z/u, ""
    gsub! /\n\n+/u, "\n\n"
    while start_with? "\n"
      slice! 0
    end
  end

  # Remove blanks preceding newlines, remove trailing
  # spaces, remove newlines succeeding empty lines, and
  # remove leading newlines.
  def trim
    copy = String.new self
    copy.trim!
    copy
  end

  # Replace space sequences by a single ASCII space.
  def slim!
    gsub! /\p{Space}+/u, " "
  end

  # Replace space sequences by a single ASCII space.
  def slim
    gsub /\p{Space}+/u, " "
  end

  # Expand tabs, subject to tabstop +stop+.
  def expand_tab! stop
    gsub! /([^\t\n]*)\t/u do
      $1 + " " * (stop - $1.size % stop)
    end
  end

  # Expand tabs, subject to tabstop +stop+.
  def expand_tab stop
    gsub /([^\t\n]*)\t/u do
      $1 + " " * (stop - $1.size % stop)
    end
  end

  # Distance from last newline to +length+.
  # If there's no last newline, just +length+.
  def find_indent
    if (offset = rindex /\n/u)
      length - offset - 1
    else
      length
    end
  end

  # Format as a C-style comment.
  def comment
    s = "/*"
    each_line do |line|
      s.concat " "
      s.concat line
      s.concat " *"
    end
    s.concat "/"
  end

  # Remove backslashes which escape characters +chars+.
  def unescape *chars
    s = String.new
    prev_char_bslash = false
    each_char do |char|
      if prev_char_bslash
        prev_char_bslash = false
        unless chars.include? char
          s.concat '\\'
        end
      else
        if char == '\\'
          prev_char_bslash = true
          next
        end
      end
      s.concat char
    end
    s
  end

  # For sentence-like strings. Truncate to +n+ characters and replace the
  # final broken word with an ellipsis.
  def truncate n
    if size <= n
      return String.new self
    else
      return String.new(self[0..n])
          .gsub(/\p{Space}+/u, " ")
          .gsub(/\p{Space}+\p{Graph}+\z/u, "...")
    end
  end

end # class String

end # module Schizm
